//
// file: primitive3D_sphere_nrm_dis_rt.fx
// version: v1.5
// author: Ren712
//

//--------------------------------------------------------------------------------------
// Settings
//--------------------------------------------------------------------------------------
float3 sElementPosition = float3(0, 0, 0);
float3 sElementRotation = float3(0, 0, 0);
float4 sElementColor = float4(0, 0, 0, 0);
float3 sElementSize = float3(1,1,1);
float2 sSubdivUnit = float2(0.08,0.08);

float sSpecularPower = 4;
float3 sSpecularColor = float3(1, 1, 1);
float3 sLightDir = float3(0.507, -0.507, -0.2);
float sOffsetBias = 0.015;
float sNormalStrength = 0.5;
float sReflectionIntensity = 0.3;
float sReflectionNormalStrength = 0.5;

bool sFlipTexture = false;
bool sFlipVertex = false;
bool sIsBillboard = false;
bool sFogEnable = false;
int fCullMode = 1;

float2 gDistFade = float2(250, 150);
float fDestBlend = 2;

float2 uvMul = float2(1, 1);
float2 uvPos = float2(0, 0);

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------
texture sTexColor;
texture sTexNormal;
texture sTexRef;

//--------------------------------------------------------------------------------------
// Variables set by MTA
//--------------------------------------------------------------------------------------
float4x4 gWorldViewProjection : WORLDVIEWPROJECTION;
float4x4 gProjection : PROJECTION;
float4x4 gView : VIEW;
float4x4 gViewInverse : VIEWINVERSE;
float3 gCameraPosition : CAMERAPOSITION;
static const float PI = 3.14159265f;
float gTime : TIME;
int gCapsMaxAnisotropy < string deviceCaps="MaxAnisotropy"; >;
int CUSTOMFLAGS < string skipUnusedParameters = "yes"; >;

//--------------------------------------------------------------------------------------
// Sampler 
//--------------------------------------------------------------------------------------
sampler2D SamplerColor = sampler_state
{
    Texture = (sTexColor);
    AddressU = Wrap;
    AddressV = Wrap;
    MipFilter = Linear;
    MaxAnisotropy = gCapsMaxAnisotropy;
    MinFilter = Anisotropic;
};

sampler2D SamplerNormal = sampler_state
{
    Texture = (sTexNormal);
    AddressU = Wrap;
    AddressV = Wrap;
    MipFilter = Linear;
    MaxAnisotropy = gCapsMaxAnisotropy;
    MinFilter = Anisotropic;
};

sampler2D SamplerReflection = sampler_state
{
    Texture = (sTexRef);
    AddressU = Wrap;
    AddressV = Wrap;
    MipFilter = Linear;
    MaxAnisotropy = gCapsMaxAnisotropy;
    MinFilter = Anisotropic;
};

//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------
struct VSInput
{
    float3 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float4 Diffuse : COLOR0;
};

struct PSInput
{
    float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float3 WorldPos : TEXCOORD1;
    float3 Normal : TEXCOORD2;
    float3 Binormal : TEXCOORD3;
    float3 Tangent : TEXCOORD4;
    float DistFade : TEXCOORD5;
    float2 Depth : TEXCOORD6;
    float4 Diffuse : COLOR0;
};

//--------------------------------------------------------------------------------------
// Create world matrix with world position and euler rotation
//--------------------------------------------------------------------------------------
float4x4 createWorldMatrix(float3 pos, float3 rot)
{
    float4x4 eleMatrix = {
        float4(cos(rot.z) * cos(rot.y) - sin(rot.z) * sin(rot.x) * sin(rot.y), 
                cos(rot.y) * sin(rot.z) + cos(rot.z) * sin(rot.x) * sin(rot.y), -cos(rot.x) * sin(rot.y), 0),
        float4(-cos(rot.x) * sin(rot.z), cos(rot.z) * cos(rot.x), sin(rot.x), 0),
        float4(cos(rot.z) * sin(rot.y) + cos(rot.y) * sin(rot.z) * sin(rot.x), sin(rot.z) * sin(rot.y) - 
                cos(rot.z) * cos(rot.y) * sin(rot.x), cos(rot.x) * cos(rot.y), 0),
        float4(pos.x,pos.y,pos.z, 1),
    };
    return eleMatrix;
}

//--------------------------------------------------------------------------------------
// Transform quad into sphere 
//--------------------------------------------------------------------------------------
float3 getSpherePosition(float3 inPosition, float3 scale)
{
    float3 outPosition;
    outPosition.z = cos(2 * inPosition.x * PI) / 2;
	outPosition.x = sin(2 * inPosition.x * PI) / 2;
	outPosition.xz *= cos(inPosition.y * PI);
	outPosition.y = sin(inPosition.y * PI) / 2;
    return outPosition * scale;
}

//--------------------------------------------------------------------------------------
// Vertex Shader 
//--------------------------------------------------------------------------------------
PSInput VertexShaderFunction(VSInput VS)
{
    PSInput PS = (PSInput)0;

    // set proper position and scale of the quad
    VS.Position.xyz = float3(- 0.5 + VS.TexCoord.xy, 0);
	
    // shape the sphere
    float3 scaleNorm = normalize(sElementSize);
    float3 resultPos = getSpherePosition(VS.Position.xyz, scaleNorm);
	
    // Calculate normals
    float3 neighbour1	= getSpherePosition(VS.Position.xyz - float3(sSubdivUnit.x, 0, 0), scaleNorm);
    float3 neighbour2	= getSpherePosition(VS.Position.xyz - float3(0, sSubdivUnit.y, 0), scaleNorm);
 
    float3 Tangent = normalize(resultPos - neighbour1);
    float3 Binormal = normalize(resultPos - neighbour2);
    float3 Normal	= cross(Tangent, Binormal);	
	
    // scale the sphere
    VS.Position.xyz = resultPos * length(sElementSize);	
	
    if (!sFlipTexture) VS.TexCoord.y = 1 - VS.TexCoord.y;
    if (sFlipVertex) VS.Position.xyz = VS.Position.xzy;
	
    // create WorldMatrix for the quad
    float4x4 sWorld = createWorldMatrix(sElementPosition, sElementRotation);

    // calculate screen position of the vertex
    float4 wPos = mul(float4( VS.Position, 1), sWorld);
    PS.WorldPos = wPos.xyz;
    float4 vPos = mul(wPos, gView);
    PS.Depth = float2(vPos.z, vPos.w);
    PS.Position = mul(vPos, gProjection);
	
    // get world normal, binormal and tangent
    PS.Normal = mul(Normal, (float3x3)sWorld).xyz;
    PS.Binormal = mul(Binormal, (float3x3)sWorld).xyz;
    PS.Tangent = mul(Tangent, (float3x3)sWorld).xyz;
	
    // get clip values
    float nearClip = - gProjection[3][2] / gProjection[2][2];
    float farClip = (gProjection[3][2] / (1 - gProjection[2][2]));
	
    // fade
    float DistFromCam = distance(gCameraPosition, sElementPosition.xyz);
    float elementSize = max(max(sElementSize.x, sElementSize.y), sElementSize.z);
    float2 DistFade = float2(min(gDistFade.x, farClip - elementSize / 2), min(gDistFade.y, farClip - elementSize /2 - (gDistFade.x - gDistFade.y)));
    PS.DistFade = saturate((DistFromCam - DistFade.x)/(DistFade.y - DistFade.x));

    // pass texCoords and vertex color to PS
    PS.TexCoord = (VS.TexCoord * uvMul) + uvPos;
    PS.Diffuse = VS.Diffuse * sElementColor;
    return PS;
}

//--------------------------------------------------------------------------------------
// getReflectionCoords
//--------------------------------------------------------------------------------------
float2 getReflectionCoords1(float3 dir)
{
    float m = 2. * sqrt(pow( dir.x, 2. ) + pow( dir.y, 2. ) + pow( dir.z + 1., 2. ));
    return dir.xy / m + .5;	
}

float2 getReflectionCoords2(float3 dir, float2 div)
{
    return float2(((atan2(dir.x, dir.z) / (PI * div.x)) + 1) / 2,  (acos(- dir.y) / (PI * div.y)));
}

//--------------------------------------------------------------------------------------
// reflectFactor
//--------------------------------------------------------------------------------------
float3 reflectFactor(float3 i, float3 n, float f)
{
    return i - 2 * n * dot(i, n) * f;
}

//--------------------------------------------------------------------------------------
// MTAApplyFog
//--------------------------------------------------------------------------------------
int gFogEnable                     < string renderState="FOGENABLE"; >;
float4 gFogColor                   < string renderState="FOGCOLOR"; >;
float gFogStart                    < string renderState="FOGSTART"; >;
float gFogEnd                      < string renderState="FOGEND"; >;

float3 MTAApplyFog( float3 texel, float distFromCam )
{
    if ( !gFogEnable )
        return texel;
    float FogAmount = ( distFromCam - gFogStart )/( gFogEnd - gFogStart );
    texel.rgb = lerp(texel.rgb, gFogColor.rgb, saturate( FogAmount) );
    return texel;
}

//--------------------------------------------------------------------------------------
// Pixel shaders 
//--------------------------------------------------------------------------------------
float4 PixelShaderFunction(PSInput PS) : COLOR0
{
    // eye to camera vector
    float3 view = (gCameraPosition - PS.WorldPos);
	
    // calculate eye vector and sample height texture for displacement effect
    float3x3 tangentToWorldSpace = float3x3(PS.Tangent, -PS.Binormal, PS.Normal);
    float3x3 worldToTangentSpace = transpose(tangentToWorldSpace);

    float3 eyeVector = normalize(mul(-view, worldToTangentSpace));
    float Height = tex2D(SamplerNormal, PS.TexCoord.xy).a;

    // change texture coords by offset value
    float2 offset = eyeVector.xy * (Height * 2.0 - 1.0) * sOffsetBias;
    float2 disTexCoord = PS.TexCoord.xy + offset;

    // sample normal texture and calculate bump normals
    float3 NormalTex = tex2D(SamplerNormal, disTexCoord).xyz;
    NormalTex.xyz = normalize((NormalTex.xyz * 2.0) - 1.0);
    float3 Normal = normalize(NormalTex.x * normalize(PS.Tangent) + NormalTex.y * normalize(PS.Binormal) + NormalTex.z * normalize(PS.Normal));
    Normal = normalize(Normal);
    Normal = lerp(PS.Normal, Normal, sNormalStrength);

    // calculate specular light
    float3 lightDir = normalize(sLightDir);
    float3 h = normalize(normalize(view) - lightDir);
    float specLighting = pow(saturate(dot(h,normalize(Normal))), sSpecularPower);
	
    // Stop underneath artifacts
    float lightAwayDot = -dot(lightDir, PS.Normal);
    specLighting *= saturate(3 * lightAwayDot);
    specLighting = saturate(specLighting);	

    // getTexCoords
    float3 eyeToCamera = normalize(view);
    eyeToCamera = reflectFactor(eyeToCamera, Normal, sReflectionNormalStrength);
    float2 refCoord = getReflectionCoords1(eyeToCamera.xzy);
	
    // get texture
    float4 texel = tex2D(SamplerColor, disTexCoord);
    float4 finalColor = texel;

    // get reflection
    float4 refTex = tex2D(SamplerReflection, refCoord);
    finalColor.rgb = lerp(texel.rgb, refTex.rgb, sReflectionIntensity);
	
    // add specular lighting to color texture
    finalColor.rgb += finalColor.rgb * specLighting * sSpecularColor.rgb;
	
    // multiply by vertex color
    finalColor *= PS.Diffuse;

    // apply distFade
    finalColor.a *= saturate( PS.DistFade );
	
    // apply fog effect
    finalColor.rgb = MTAApplyFog(finalColor.rgb, PS.Depth.x / PS.Depth.y);

    return saturate(finalColor);
}

float4 PixelShaderFunctionSM2(PSInput PS) : COLOR0
{
    // eye to camera vector
    float3 view = (gCameraPosition - PS.WorldPos);
	
    // calculate eye vector and sample height texture for displacement effect
    float3x3 tangentToWorldSpace = float3x3(PS.Tangent, -PS.Binormal, PS.Normal);
    float3x3 worldToTangentSpace = transpose(tangentToWorldSpace);

    float3 eyeVector = normalize(mul(-view, worldToTangentSpace));
    float Height = tex2D(SamplerNormal, PS.TexCoord.xy).a;

    // change texture coords by offset value
    float2 offset = eyeVector.xy * (Height * 2.0 - 1.0) * sOffsetBias;
    float2 disTexCoord = PS.TexCoord.xy + offset;

    // sample normal texture and calculate bump normals
    float3 NormalTex = tex2D(SamplerNormal, disTexCoord).xyz;
    NormalTex.xyz = normalize((NormalTex.xyz * 2.0) - 1.0);
    float3 Normal = normalize(NormalTex.x * normalize(PS.Tangent) + NormalTex.y * normalize(PS.Binormal) + NormalTex.z * normalize(PS.Normal));
    Normal = normalize(Normal);
    Normal = lerp(PS.Normal, Normal, sNormalStrength);

    // calculate specular light
    float3 lightDir = normalize(sLightDir);
    float3 h = normalize(normalize(view) - lightDir);
    float specLighting = pow(saturate(dot(h,normalize(Normal))), sSpecularPower);
	
    // Stop underneath artifacts
    float lightAwayDot = -dot(lightDir, PS.Normal);
    specLighting *= saturate(3 * lightAwayDot);
    specLighting = saturate(specLighting);	
	
    // get texture
    float4 texel = tex2D(SamplerColor, disTexCoord);
    float4 finalColor = texel;
	
    // add specular lighting to color texture
    finalColor.rgb += finalColor.rgb * specLighting * sSpecularColor.rgb;
	
    // multiply by vertex color
    finalColor *= PS.Diffuse;

    // apply distFade
    finalColor.a *= saturate( PS.DistFade );

    return saturate(finalColor);
}

//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------
technique dxDrawPrimitive3Dfx_lightBall
{
  pass P0
  {
    ZEnable = true;
    ZFunc = LessEqual;
    ZWriteEnable = true;
    CullMode = fCullMode;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = fDestBlend;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_3_0 VertexShaderFunction();
    PixelShader  = compile ps_3_0 PixelShaderFunction();
  }
}

technique dxDrawPrimitive3Dfx_lightBall_SM2
{
  pass P0
  {
    ZEnable = true;
    ZFunc = LessEqual;
    ZWriteEnable = true;
    CullMode = fCullMode;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = fDestBlend;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction();
    PixelShader  = compile ps_2_0 PixelShaderFunctionSM2();
  }
}

// Fallback
technique fallback
{
  pass P0
  {
    // Just draw normally
  }
}
